# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import logging
from typing import Callable, Dict

import numpy as np
import torch
from torch.fx import symbolic_trace
from torch.fx.passes.shape_prop import ShapeProp

from model_compression_toolkit.common import Graph
from model_compression_toolkit.pytorch.reader.graph_builders import edges_builder, nodes_builder
from model_compression_toolkit.pytorch.utils import set_model


def generate_module_dict(model: torch.nn.Module) -> Dict:
    """
    Creates a dictionary from the Pytorch model's named modules by module name.

    Args:
        model: A Pytorch model.

    Returns:
        A dictionary of the Pyotrch model's named modules.
    """
    module_dict = dict()
    for name, module in model.named_modules():
        module_dict[name] = module
    return module_dict


def build_graph(model: torch.fx.GraphModule,
                to_numpy: Callable) -> Graph:
    """
    Given a Pytorch FX model, build and return an networkx MultiDiGraph containing all data (nodes, edges,
    inputs and outputs) representing that model.

    Args:
        model: Pytorch FX model to build its graph.
        to_numpy: Function to convert framework's tensor to a Numpy array.

    Returns:
        Networkx MultiDiGraph representing the Pytorch model.

    """
    # generate a dictionary with all model modules
    module_dict = generate_module_dict(model)

    # convert fx nodes to generic graph nodes
    nodes, inputs, outputs, fx_node_2_graph_node = nodes_builder(model, module_dict, to_numpy)

    # build graph edges
    edges = edges_builder(model, fx_node_2_graph_node)

    return Graph(nodes, inputs, outputs, edges)


def fx_graph_module_generation(pytorch_model: torch.nn.Module,
                               representative_data_gen: Callable,
                               to_tensor: Callable) -> torch.fx.GraphModule:
    """
    Generates a fx.GraphModule from a torch.nn.Module.

    Args:
        pytorch_model: A dynamic Pytorch model.
        representative_data_gen (Callable): Representative dataset used for shape inference.
        to_tensor: Function to convert a Numpy array to a framework's tensor.

    Returns:
        A fx.GraphModule (static model) representing the Pytorch model.
    """
    set_model(pytorch_model)
    symbolic_traced = symbolic_trace(pytorch_model)
    inputs = representative_data_gen()
    input_for_shape_infer = [to_tensor(i) for i in inputs]
    ShapeProp(symbolic_traced).propagate(*input_for_shape_infer)
    return symbolic_traced


def model_reader(model: torch.nn.Module,
                 representative_data_gen: Callable,
                 to_numpy: Callable,
                 to_tensor: Callable) -> Graph:
    """
    Reads a Pytorch model, converts it to an FX Graph using the fx toolkit, then builds a base graph representing the fx graph.
    Args:
        model: Pytorch model to build its graph representation.
        representative_data_gen (Callable): Dataset used for calibration.
        to_numpy: Function to convert framework's tensor to a Numpy array.
        to_tensor: Function to convert a Numpy array to a framework's tensor.

    Returns:
        Base graph of the Pytorch model.
    """
    logging.info("Start Model Reading...")
    fx_model = fx_graph_module_generation(model, representative_data_gen, to_tensor)
    graph = build_graph(fx_model, to_numpy)
    return graph


# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from enum import Enum
from typing import List, Callable

from model_compression_toolkit.common.mixed_precision.distance_weighting import get_average_weights
from model_compression_toolkit.common.quantization.quantization_config import QuantizationConfig, DEFAULTCONFIG
from model_compression_toolkit.common.similarity_analyzer import compute_mse


class MixedPrecisionQuantizationConfig(QuantizationConfig):

    def __init__(self,
                 qc: QuantizationConfig = DEFAULTCONFIG,
                 weights_n_bits: List[int] = None,
                 compute_distance_fn: Callable = compute_mse,
                 distance_weighting_method: Callable = get_average_weights,
                 num_of_images: int = 32,
                 configuration_overwrite: List[int] = None):
        """
        Class to wrap all different parameters the library quantize the input model according to.
        Unlike QuantizationConfig, number of bits for quantization is a list of possible bit widths to
        support mixed-precision model quantization.

        Args:
            qc (QuantizationConfig): QuantizationConfig object containing parameters of how the model should be quantized.
            weights_n_bits (List[int]): List of possible number of bits to quantize the coefficients.
            compute_distance_fn (Callable): Function to compute a distance between two tensors.
            distance_weighting_method (Callable): Function to use when weighting the distances among different layers when computing the sensitivity metric.
            num_of_images (int): Number of images to use to evaluate the sensitivity of a mixed-precision model comparing to the float model.
            configuration_overwrite (List[int]): A list of integers that enables overwrite of mixed precision with a predefined one.

        """

        super().__init__(**qc.__dict__)
        self.weights_n_bits = weights_n_bits if weights_n_bits is not None else [qc.weights_n_bits]
        self.compute_distance_fn = compute_distance_fn
        self.distance_weighting_method = distance_weighting_method
        self.num_of_images = num_of_images
        self.configuration_overwrite = configuration_overwrite


# Default quantization configuration the library use.
DEFAULT_MIXEDPRECISION_CONFIG = MixedPrecisionQuantizationConfig(DEFAULTCONFIG,
                                                                 [2, 4, 8],
                                                                 compute_mse,
                                                                 get_average_weights)



from typing import Any, Tuple

import tensorflow as tf
if tf.__version__ < "2.6":
    from tensorflow.keras.layers import Activation, ReLU
else:
    from keras.layers import Activation, ReLU

from model_compression_toolkit import FrameworkInfo
from model_compression_toolkit.common import BaseNode
from model_compression_toolkit.common.node_prior_info import NodePriorInfo
from model_compression_toolkit.keras.constants import ACTIVATION, RELU_MAX_VALUE, NEGATIVE_SLOPE, THRESHOLD


def create_node_prior_info(node: BaseNode,
                           fw_info: FrameworkInfo):
    """
    Create a NodePriorInfo object for a given node.

    Args:
        node: Node to create its prior info.
        fw_info: Information about a specific framework the node was generated from.

    Returns:
        NodePriorInfo object with info about the node.
    """

    min_output, max_output = _get_min_max_outputs(node=node,
                                                  fw_info=fw_info)
    return NodePriorInfo(min_output=min_output,
                         max_output=max_output)


def _get_min_max_outputs(node: BaseNode,
                         fw_info: FrameworkInfo) -> Tuple[Any,Any]:
    """
    Return the min/max output values of a node if known.
    If one of them (or both of them) is unknown - return None instead of a value.
    Args:
        node: Node to create its prior info.
        fw_info: Information about a specific framework the node was generated from.

    Returns:
        Min/max output values if known.
    """
    min_output, max_output = None, None

    if node.type == ReLU:
        min_output = node.framework_attr[THRESHOLD] if node.framework_attr[NEGATIVE_SLOPE] == 0 else None
        max_output = node.framework_attr[RELU_MAX_VALUE]

    elif fw_info.layers_has_min_max(node.type):
        min_output, max_output = fw_info.layer_min_max_mapping[node.type]

    elif node.type == Activation and fw_info.activation_has_min_max(node.framework_attr[ACTIVATION]):
        min_output, max_output = fw_info.activation_min_max_mapping[node.framework_attr[ACTIVATION]]

    return min_output, max_output





# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


# Minimal threshold to use for quantization ranges:
MIN_THRESHOLD = (2 ** -28)
EPS = 1e-8
MULTIPLIER_N_BITS = 8

# Quantization attributes:
OUTPUT_SCALE = 'output_scale'
THRESHOLD = 'threshold'
SIGNED = 'is_signed'
CLUSTER_CENTERS = 'cluster_centers'
SCALE_PER_CHANNEL = 'scale_per_channel'
RANGE_MIN = 'range_min'
RANGE_MAX = 'range_max'


# Data types:
DATA_TYPE = 'dtype'
FLOAT_32 = 'float32'


# Number of Tensorboard cosine-similarity plots to add:
NUM_SAMPLES_CS_TENSORBOARD = 20

# num bits for shift negative non linear node
SHIFT_NEGATIVE_NON_LINEAR_NUM_BITS = 16

# In Mixed-Precision, a node can have multiple candidates for weights quantization configuration.
# In order to display a single view of a node (for example, for logging in TensorBoard) we need to track the attributes
# that are shared among different candidates:
WEIGHTS_NBITS_ATTRIBUTE = 'weights_n_bits'
CORRECTED_BIAS_ATTRIBUTE = 'corrected_bias'





# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import tensorflow as tf
# As from Tensorflow 2.6, keras is a separate package and some classes should be imported differently.
if tf.__version__ < "2.6":
    from tensorflow.python.keras.engine.node import Node as KerasNode
else:
    from keras.engine.node import Node as KerasNode


from tensorflow.python.util.object_identity import Reference as TFReference
from typing import List, Tuple

from model_compression_toolkit.common.graph.base_graph import OutTensor
from model_compression_toolkit.common.graph.base_node import BaseNode
from model_compression_toolkit.keras.reader.common import is_node_an_input_layer
from model_compression_toolkit.keras.reader.node_builder import build_node

keras = tf.keras
layers = keras.layers


class ConnectivityHandler(object):
    """
    Class containing all information about connections in a graph: nodes and input/output tensors of each node.
    A connection between two nodes is exist if there is a tensor connecting them.
    Tensors are represented by their names.
    """

    def __init__(self):
        self._nodes2input_tensors = dict()  # Node -> List[Tensor]
        self._input_tensors2nodes = dict()  # Tensor -> List[Node]
        self._nodes2output_tensors = dict()  # Node -> List[Tensor]
        self._output_tensors2nodes = dict()  # Tensor -> Node

    def get_nodes(self) -> List[BaseNode]:
        """
        Returns: List of nodes in the connectivity handler.
        """
        return list(self._nodes2input_tensors.keys())

    def is_tensor_connected(self,
                            tensor: str) -> bool:
        """
        Whether a tensor is an input tensor of a node in the connectivity handler.
        Args:
            tensor: Tensor reference to check whether is an input tensor of a node in the connectivity handler.

        Returns:
            Whether a tensor is an input tensor of a node in the connectivity handler.
        """
        return self._input_tensors2nodes.get(tensor) is not None

    def input_tensor2nodes(self,
                           in_tensor: str) -> List[BaseNode]:
        """
        Returns a list of nodes that have a given tensor in their input tensors.
        Args:
            in_tensor: Tensor's reference which the returning nodes list hold in their input tensors.

        Returns:
            List of nodes that have in_tensor in their input tensors.
        """
        return self._input_tensors2nodes[in_tensor] if in_tensor in self._input_tensors2nodes else []

    def output_tensor2node(self,
                           out_tensor: str) -> KerasNode:
        """
        Returns the node that its output is the tensor reference out_tensor.
        Args:
            out_tensor: Tensor reference which the returning node holds in their output tensors.

        Returns:
            A node that has out_tensor in its output tensors.
        """
        return self._output_tensors2nodes[out_tensor] if out_tensor in self._output_tensors2nodes else None

    def node2input_tensors(self,
                           node: BaseNode) -> List[TFReference]:
        """
        Get a list of input tensors of a node.
        Args:
            node: Node to return its input tensors.

        Returns:
            List of references to input tensors of the given node.
        """
        return self._nodes2input_tensors[node] if node in self._nodes2input_tensors else []

    def node2output_tensors(self,
                            node: BaseNode) -> List[TFReference]:
        """
        Get a list of output tensors of a node.
        Args:
            node: Node to return its output tensors.

        Returns:
            List of references to output tensors of the given node.
        """
        return self._nodes2output_tensors[node] if node in self._nodes2output_tensors else []

    def build_inputs_list(self,
                          model_input_tensors_list: List[TFReference]) -> List[KerasNode]:
        """
        Build a list of input nodes of the model ordered by their model's input order.

        Args:
            model_input_tensors_list: List of references to the model's input tensors.

        Returns:
            List of the model input nodes ordered by their indices.
        """
        input_nodes_list = []
        for it in model_input_tensors_list:  # iterate model inputs list
            for n in self._input_tensors2nodes[it]:
                if is_node_an_input_layer(n):  # verify it is an input layer
                    input_nodes_list.append(n)
        return input_nodes_list

    def build_outputs_list(self,
                           model_output_tensors_list: List[TFReference]) -> List[OutTensor]:
        """
        Build a list of model output nodes. Each node is being mapped to a list of two integers tuples. Each tuple
        contains the information about indices a model output tensor gets. The first index in the tuple is the index
        of the tensor
        relatively to the node output tensors. The second index in the tuple is the index of the tensor
        relatively to the model output tensors.
        Args:
            model_output_tensors_list:  List of references to the model's output tensors.

        Returns:
            List of the model output nodes and their output indices.
        """
        outputs_list = []
        for model_ot in model_output_tensors_list:
            out_node = self._output_tensors2nodes.get(model_ot)  # output node of the model
            # index of tensor relatively to the node output tensors
            node_out_index = self._nodes2output_tensors.get(out_node).index(model_ot)
            outputs_list.append(OutTensor(out_node, node_out_index))
        return outputs_list

    def add_node(self,
                 node: KerasNode,
                 input_tensors_list: List[TFReference],
                 outputs_tensors_list: List[TFReference]):
        """
        Add a node and its input/output tensors to the connectivity handler.
        Args:
            node: Node in the graph.
            input_tensors_list: Node's input tensors references.
            outputs_tensors_list: Node's output tensors references.

        """
        self._nodes2input_tensors[node] = input_tensors_list
        self._nodes2output_tensors[node] = outputs_tensors_list
        for input_t in input_tensors_list:  # update dictionary from a tensor to all nodes having this tensor in
            # their input tensors
            if self._input_tensors2nodes.get(input_t) is None:
                self._input_tensors2nodes[input_t] = [node]
            else:
                self._input_tensors2nodes[input_t].append(node)
        # update dictionary from a tensor to the node its is output tensor.
        # There is only one node that can output this tensor, thus it's not a list
        for output_t in outputs_tensors_list:
            self._output_tensors2nodes[output_t] = node

    def get_edge_indices(self,
                         src_node: BaseNode,
                         dst_node: BaseNode,
                         connecting_tensor: TFReference) -> Tuple[int, int]:
        """
        Get indices of an edge by its source/destination nodes and the connecting tensor which defines the edge.
        The indices are the relative positions of the connecting tensor regard input/output tensors of
        destination/source nodes.
        Args:
            src_node: Source node of the edge.
            dst_node: Destination node of the edge.
            connecting_tensor: Tensor reference that connects the source and destination nodes.

        Returns:
            Tuple of two edge indices: source index and destination index.
        """
        src_index = self.node2output_tensors(src_node).index(connecting_tensor)
        dst_index = self.node2input_tensors(dst_node).index(connecting_tensor)
        return src_index, dst_index

    def get_out_edges_params_list(self,
                                  src_node: BaseNode) -> List[tuple]:
        """
        Compute for a given node, all parameters of its outgoing edges.
        Args:
            src_node: Node to consider as source node to compute outgoing edges parameters.

        Returns:
            List of tuples. Each tuple contains parameters of the outgoing edge of src_node: destination node,
            source index, destination index.
        """
        out_edges_params_list = []
        out_tensors = self.node2output_tensors(src_node)
        for ot in out_tensors:  # Loop over output tensors
            if self.is_tensor_connected(ot):
                for dst_node in self.input_tensor2nodes(ot):  # Loop over dst nodes
                    if src_node is not dst_node:  # To handle input layer due to equality of input and output tensors
                        # Get edge indices for the connecting tensor
                        src_index, dst_index = self.get_edge_indices(src_node, dst_node, ot)
                        out_edges_params_list.append((dst_node, src_index, dst_index))

        return out_edges_params_list

    def convert_to_internal_nodes(self):
        """
        Convert nodes in the connectivity handler from a Keras node to an internal node.
        Each internal node is created from a Keras node the connectivity handler has.
        """

        keras_nodes = self.get_nodes()
        node_name_to_node = dict()  # dictionary from node name to node for handling reused layers
        keras_node_to_internal_node = dict()
        for node in keras_nodes:
            internal_node = build_node(node, node_name_to_node)  # build node in the graph
            node_name_to_node[internal_node.name] = internal_node  # update nodes dictionary
            keras_node_to_internal_node[node] = internal_node

        # update _nodes2input_tensors and _nodes2output_tensors with new converted nodes
        for keras_node, internal_node in keras_node_to_internal_node.items():
            self._nodes2input_tensors[internal_node] = self.node2input_tensors(keras_node)
            self._nodes2input_tensors.pop(keras_node)
            self._nodes2output_tensors[internal_node] = self.node2output_tensors(keras_node)
            self._nodes2output_tensors.pop(keras_node)

        # update _input_tensors2nodes
        for it, keras_nodes_list in self._input_tensors2nodes.items():
            self._input_tensors2nodes[it] = [keras_node_to_internal_node[k_node] for k_node in keras_nodes_list]

        # update _output_tensors2nodes
        for ot, mapped_keras_node in self._output_tensors2nodes.items():
            self._output_tensors2nodes[ot] = keras_node_to_internal_node[mapped_keras_node]



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from model_compression_toolkit import FrameworkInfo
from model_compression_toolkit.common import BaseNode
from model_compression_toolkit.common.node_prior_info import NodePriorInfo


def create_node_prior_info(node: BaseNode,
                           fw_info: FrameworkInfo):
    """
    Create a NodePriorInfo object for a given node.

    Args:
        node: Node to create its prior info.
        fw_info: Information about a specific framework the node was generated from.

    Returns:
        NodePriorInfo object with info about the node.
    """

    min_output, max_output = None, None
    if fw_info.layers_has_min_max(node.type):
        min_output, max_output = fw_info.layer_min_max_mapping[node.type]
    return NodePriorInfo(min_output=min_output,
                         max_output=max_output)



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from collections import Callable
from typing import Any
from scipy import optimize

import numpy as np

from model_compression_toolkit.common.constants import MIN_THRESHOLD, THRESHOLD
from model_compression_toolkit.common.quantization.quantizers.quantizers_helpers import quantize_tensor, \
    reshape_tensor_for_per_channel_search, uniform_quantize_tensor, get_output_shape, get_range_bounds, \
    get_threshold_bounds
from model_compression_toolkit.common.quantization.quantization_params_generation.no_clipping import \
    no_clipping_selection_tensor, no_clipping_selection_histogram


def qparams_selection_tensor_search(error_function: Callable,
                                    tensor_data: np.ndarray,
                                    n_bits: int,
                                    per_channel: bool = False,
                                    channel_axis: int = 1,
                                    n_iter: int = 10,
                                    min_threshold=MIN_THRESHOLD) -> Any:
    """
    Search for an optimal threshold to quantize a tensor.
    The search_methods starts with the constrained no-clipping threshold the tensor has, and continues with
    n_iter another smaller constrained thresholds. For each candidate threshold, an error is computed
    based on the passed error function, and the threshold which yields the minimal error is selected
    and returned.

    Args:
        error_function: Function to compute the error between the original and quantized tensors.
        tensor_data: Numpy array with tensor's content.
        n_bits: Number of bits to quantize the tensor.
        per_channel: Whether the tensor should be quantized per-channel or per-tensor.
        channel_axis: Index of output channels dimension.
        n_iter: Number of searching iterations.
        min_threshold: Threshold to return if the computed threshold is smaller that min_threshold.

    Returns:
        Optimal constrained threshold to quantize the tensor.

    """

    signed = np.any(tensor_data < 0)  # check if tensor is singed
    output_shape = get_output_shape(tensor_data.shape, channel_axis)

    # First threshold to check is the constrained threshold based on the tensor's maximal value.
    threshold = 2 * no_clipping_selection_tensor(tensor_data,
                                                 0,
                                                 n_bits,
                                                 per_channel,
                                                 channel_axis,
                                                 min_threshold=min_threshold)[THRESHOLD]

    # If the threshold is computed per-channel, we rearrange the tensor such that each sub-tensor
    # is flattened, and we iterate over each one of them when searching for the threshold.
    if per_channel:
        tensor_data_r = reshape_tensor_for_per_channel_search(tensor_data, channel_axis)

    error_list = []  # init an empty error list
    # On each iteration a new constrained threshold which equal to half of the previous tested threshold
    # is used for quantizing the tensor and computing the error. The error is appended to an error list, which
    # eventually used to select the threshold with the minimal error.
    for i in range(n_iter):
        if per_channel:
            per_channel_error = []
            for j in range(tensor_data_r.shape[0]):  # iterate all channels of the tensor.
                qt = quantize_tensor(tensor_data_r[j, :], threshold.flatten()[j] / (2 ** i), n_bits, signed)
                error = error_function(qt, tensor_data_r[j, :], threshold=threshold.flatten()[j] / (2 ** i))
                per_channel_error.append(error)
            error_list.append(np.asarray(per_channel_error))
        else:  # quantize per-tensor
            qt = quantize_tensor(tensor_data, threshold / (2 ** i), n_bits, signed)
            error = error_function(qt, tensor_data, threshold=threshold / (2 ** i))
            error_list.append(error)

    # Take the index of the minimal error, and use it compute the threshold which yielded it.
    i = np.argmin(np.stack(error_list, axis=-1), axis=-1)

    return np.maximum(np.reshape(threshold.flatten() / np.power(2, i), output_shape), min_threshold)


def qparams_selection_histogram_search(error_function: Callable,
                                       bins: np.ndarray,
                                       counts: np.ndarray,
                                       n_bits: int,
                                       constrained: bool = True,
                                       n_iter: int = 10,
                                       min_threshold: float = MIN_THRESHOLD):
    """
    Search for an optimal threshold to quantize a histogram of collected float values.
    The search_methods starts with the constrained no-clipping threshold by the bins' maximal value, and continues with
    n_iter another smaller constrained thresholds. For each candidate threshold, an error is computed
    based on the passed error function, and the threshold which yields the minimal error is selected
    and returned.

    Args:
        error_function: Function to compute the error between the original and quantized histograms.
        bins: Bins of the histogram to search_methods for an optimal threshold.
        counts: Number of elements in the bins to search_methods for a threshold.
        n_bits: Number of bits to quantize the tensor.
        constrained: Whether the threshold should be constrained or not.
        n_iter: Number of searching iterations.
        min_threshold: Threshold to return if the computed threshold is smaller that min_threshold.

    Returns:
        Optimal constrained threshold to quantize the tensor.

    """

    signed = np.any(bins < 0)  # Whether histogram contains negative values or not.
    threshold = (1 + int(constrained)) * no_clipping_selection_histogram(bins,
                                                                         counts,
                                                                         p=0,#dummy
                                                                         n_bits=n_bits,#dummy
                                                                         min_value=0,#dummy
                                                                         max_value=0,#dummy
                                                                         constrained=constrained,
                                                                         n_iter=n_iter, #dummy
                                                                         min_threshold=min_threshold)
    # Init a list of thresholds.
    error_list = []
    threshold_list = threshold / np.power(2, np.linspace(0, n_iter - 1, n_iter))

    # On each iteration a new constrained threshold which equal to half of the previous tested threshold
    # is used for quantizing the histogram and computing the error. The error is appended to an error list, which
    # eventually used to select the threshold with the minimal error.
    for threshold in threshold_list:
        q_bins = quantize_tensor(bins, threshold, n_bits, signed)  # compute the quantized values of the bins.
        error = qparams_selection_histogram_search_error_function(error_function, bins, q_bins, counts, threshold=threshold)
        error_list.append(error)

    # Return the threshold with the minimal error.
    return np.maximum(threshold_list[np.argmin(error_list)], min_threshold)


def qparams_tensor_minimization(x, x0, error_function, quant_function, bounds=None):
    """
        Search for an optimal quantization parameters to quantize a tensor.
        Uses scipy.optimization.minimize method to search through the space of possible parameters
        to quantize the tensor according to the given quantization function and possible parameters bounds.
        Used for parameters' selection search for Symmetric and Uniform quantization
        (depends on quant_function argument).

        Args:
            x: Numpy array with tensor's content.
            x0: An initial solution guess for the minimization process.
            error_function: Function to compute the error between the original and quantized tensors.
            quant_function: Function to quantize the tensor.
            bounds: Lower and upper bounds for the quantization parameters' values
            (threshold for symmetric quantization and range for uniform quantization).

        Returns:
            Optimal quantization range to quantize the tensor.

        """
    return optimize.minimize(fun=lambda qparam: error_function(x, quant_function(qparam), qparam),
                             x0=x0,
                             bounds=bounds,
                             method='Nelder-Mead')


def qparams_histogram_minimization(x, x0, counts, error_function, quant_function, bounds=None):
    """
        Search for an optimal parameters to quantize a histogram.
        Uses scipy.optimization.minimize method to search through the space of possible
        parameters for quantization according to the given quantization method and possible parameters bounds.
        Used for parameters' selection search for Symmetric and Uniform quantization
        (depends on quant_function argument).

        Args:
            x: Numpy array with tensor's content.
            x0: An initial solution guess for the minimization process.
            counts: Number of elements in the bins to search_methods for a threshold.
            error_function: Function to compute the error between the original and quantized tensors.
            quant_function: Function to quantize the tensor.
            bounds: Lower and upper bounds for the quantization parameters' values
            (threshold for symmetric quantization and range for uniform quantization).


        Returns:
            OptimizeResult object containing the optimal parameters to quantize the histogram by.

    """
    return optimize.minimize(fun=lambda qparam:
                             qparams_selection_histogram_search_error_function(error_function=error_function,
                                                                               bins=x,
                                                                               q_bins=quant_function(qparam),
                                                                               counts=counts,
                                                                               min_max_range=qparam),
                             x0=x0,
                             bounds=bounds,
                             method='Nelder-Mead')


def kl_symmetric_qparams_histogram_minimization(x, x0, counts, n_bits, signed, error_function, bounds):
    """
    Search for an optimal threshold to quantize a histogram, using the KL error method.
    Uses scipy.optimization.minimize method to search through the space of possible
    parameters for symmetric quantization.

    Args:
        x: Numpy array with tensor's content.
        x0: An initial solution guess for the minimization process.
        counts: Number of elements in the bins to search_methods for a threshold.
        n_bits: Number of bits to quantize the tensor.
        signed: Whether the quantization range should include negative values or not.
        error_function: Function to compute the error between the original and quantized tensors.
        bounds: Lower and upper bounds for the quantization parameters' values
            (threshold for symmetric quantization and range for uniform quantization).


    Returns:
        OptimizeResult object containing the optimal threshold to quantize the histogram by.

    """
    return optimize.minimize(fun=lambda threshold:
                             kl_qparams_selection_histogram_search_error_function(error_function=error_function,
                                                                                  bins=x,
                                                                                  q_bins=quantize_tensor(x,
                                                                                                         threshold,
                                                                                                         n_bits,
                                                                                                         signed),
                                                                                  counts=counts,
                                                                                  min_max_range=np.array([0, threshold]) if not signed else np.array([-threshold, threshold])),
                             x0=x0,
                             bounds=bounds,
                             method='Nelder-Mead')


def kl_uniform_qparams_histogram_minimization(x, x0, counts, n_bits, error_function, bounds=None):
    """
    Search for an optimal range to quantize a histogram, using the KL error method.
    Uses scipy.optimization.minimize method to search through the space of possible
    parameters for uniform quantization.

    Args:
        x: Numpy array with tensor's content.
        x0: An initial solution guess for the minimization process.
        counts: Number of elements in the bins to search_methods for a threshold.
        n_bits: Number of bits to quantize the tensor.
        error_function: Function to compute the error between the original and quantized tensors.
        bounds: Lower and upper bounds for the quantization parameters' values
            (threshold for symmetric quantization and range for uniform quantization).


    Returns:
        OptimizeResult object containing the optimal threshold to quantize the histogram by.

    """
    return optimize.minimize(fun=lambda min_max_range:
                             kl_qparams_selection_histogram_search_error_function(error_function=error_function,
                                                                                  bins=x,
                                                                                  q_bins=uniform_quantize_tensor(x,
                                                                                                                 range_min=min_max_range[0],
                                                                                                                 range_max=min_max_range[1],
                                                                                                                 n_bits=n_bits),
                                                                                  counts=counts,
                                                                                  min_max_range=min_max_range),
                             x0=x0,
                             bounds=bounds,
                             method='Nelder-Mead')


def qparams_selection_histogram_search_error_function(error_function: Callable,
                                                      bins: np.ndarray,
                                                      q_bins: np.ndarray,
                                                      counts: np.ndarray,
                                                      threshold: np.ndarray = None,
                                                      min_max_range=None):
    """
    Computes the error according to the given error function, to be used in the parameters' selection process
    for quantization.
    Args:
        error_function: Function to compute the error between the original and quantized histograms.
        bins: Bins values of the histogram.
        q_bins: Bins values of the quantized histogram.
        counts: Bins counts of the original histogram.
        threshold: Threshold bins were quantized by (used only for kl error function).
        min_max_range: quantization parameter, used in uniform parameters' selection for quantization range validation
        (not used for symmetric parameters' selection)

    Returns: the error between the original and quantized histogram.

    """
    # computes the number of elements between quantized bin values.
    q_count, _ = np.histogram(q_bins, bins=bins, weights=np.concatenate([counts.flatten(), np.asarray([0])]))
    # threshold is only used for KL error method calculations.
    # other error methods are passed with a wrapper that accepts a threshold argument but does not use it.
    error = error_function(q_bins, q_count, bins, counts, threshold, min_max_range)  # computes the error
    return error


def kl_qparams_selection_histogram_search_error_function(error_function: Callable,
                                                         bins: np.ndarray,
                                                         q_bins: np.ndarray,
                                                         counts: np.ndarray,
                                                         min_max_range: np.ndarray):
    """
    Computes the error according to the KL-divergence the distributions of the given histogram.
    The error value is used in the threshold optimization process, for symmetric quantization.
    Args:
        error_function: Function to compute the error between the original and quantized histograms.
        bins: Bins values of the histogram.
        q_bins: Quantized bins values of the histogram.
        counts: Bins counts of the original histogram.
        min_max_range: Quantization range to quantize histogram by.

    Returns: the error between the original and quantized histogram.

    """
    # compute the number of elements between quantized bin values.
    q_count, _ = np.histogram(q_bins, bins=bins, weights=np.concatenate([counts.flatten(), np.asarray([0])]))
    error = error_function(q_bins, q_count, bins, counts, range_min=min_max_range[0], range_max=min_max_range[1])
    return error


def symmetric_qparams_selection_per_channel_search(tensor_data, tensor_max, channel_axis, search_function, min_threshold):
    """
    A wrapper function for running a search for optimal threshold per-channel of a tensor,
    to be used in symmetric quantization.

    Args:
        tensor_data: Numpy array with tensor's content.
        tensor_max: The maximal values per channel.
        channel_axis: Index of output channels dimension.
        search_function: Function to preform the parameters' optimization for a given channel.
        min_threshold: Threshold to return if the computed threshold is smaller that min_threshold.

    Returns: ndarray with the optimal threshold for each channel, shaped according to the channels_axis.

    """
    tensor_data_r = reshape_tensor_for_per_channel_search(tensor_data, channel_axis)
    output_shape = get_output_shape(tensor_data.shape, channel_axis)
    res = []

    for j in range(tensor_data_r.shape[0]):  # iterate all channels of the tensor.
        channel_data = tensor_data_r[j, :]
        channel_threshold = max(min_threshold, tensor_max.flatten()[j])

        bounds = get_threshold_bounds(min_threshold, channel_threshold)
        channel_res = search_function(channel_data, channel_threshold, bounds)
        res.append(channel_res.x)
    return np.reshape(np.array(res), output_shape)


def uniform_qparams_selection_per_channel_search(tensor_data, tensor_min, tensor_max, channel_axis, search_function):
    """
    A wrapper function for running a search for optimal range per channel of a tensor,
    to be used in uniform quantization.

    Args:
        tensor_data: Numpy array with tensor's content.
        tensor_min: The minimal values per channel.
        tensor_max: The maximal values per channel.
        channel_axis: Index of output channels dimension.
        search_function: Function to preform the parameters' optimization for a given channel.

    Returns: ndarray with the optimal threshold for each channel, shaped according to the channels_axis.

    """
    tensor_data_r = reshape_tensor_for_per_channel_search(tensor_data, channel_axis)
    output_shape = get_output_shape(tensor_data.shape, channel_axis)
    res_min, res_max = [], []

    for j in range(tensor_data_r.shape[0]):  # iterate all channels of the tensor.
        channel_data = tensor_data_r[j, :]
        channel_range_min = tensor_min.flatten()[j]
        channel_range_max = tensor_max.flatten()[j]
        channel_min_max = np.array([channel_range_min, channel_range_max])

        bounds = get_range_bounds(channel_range_min, channel_range_max)
        channel_res = search_function(channel_data, channel_min_max, bounds)

        res_min.append(channel_res.x[0])
        res_max.append(channel_res.x[1])

    res_min = np.reshape(np.array(res_min), output_shape)
    res_max = np.reshape(np.array(res_max), output_shape)
    return res_min, res_max



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import copy
from typing import List

from model_compression_toolkit.common import Logger, BaseNode
from model_compression_toolkit.common.framework_info import FrameworkInfo
from model_compression_toolkit.common.graph.base_graph import Graph
from model_compression_toolkit.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfig
from model_compression_toolkit.common.quantization.node_quantization_config import NodeActivationQuantizationConfig, \
    NodeWeightsQuantizationConfig
from model_compression_toolkit.common.quantization.quantization_config import QuantizationConfig
from model_compression_toolkit.common.quantization.quantization_params_fn_selection import \
    get_activation_quantization_params_fn, get_weights_quantization_params_fn


def set_quantization_configuration_to_graph(graph: Graph,
                                            quant_config: QuantizationConfig,
                                            fw_info: FrameworkInfo) -> Graph:
    """
    Add quantization configuration for each graph node.

    Args:
        graph: Graph for which to add quantization info to each node.
        quant_config: Quantization configuration containing parameters for how the graph should be quantized.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
        groups of layers by how they should be quantized, etc.)

    Returns:
        The graph with quantization configurations attached to each node in it.
    """

    graph_with_qcs = copy.deepcopy(graph)
    for n in graph_with_qcs.nodes:
        set_quantization_configs_to_node(node=n,
                                         quant_config=quant_config,
                                         fw_info=fw_info)
    return graph_with_qcs


def set_quantization_configs_to_node(node: BaseNode,
                                     quant_config: QuantizationConfig,
                                     fw_info: FrameworkInfo):
    """
    Create and set quantization configurations to a node (for both weights and activation).

    Args:
        node: Node to set its quantization configurations.
        quant_config: Quantization configuration to generate the node's configurations from.
        fw_info: Information needed for quantization about the specific framework.

    """
    # Create activation QC for this node
    node.activation_quantization_cfg = create_node_activation_qc(quant_config,
                                                                 fw_info)

    enable_activation_quantization = quant_config.enable_activation_quantization and (fw_info.in_activation_ops(node) or fw_info.in_kernel_ops(node))
    node.activation_quantization_cfg.enable_activation_quantization = enable_activation_quantization

    # Create weights QC for this node
    weight_channel_axis = fw_info.kernel_channels_mapping.get(node.type)[0]
    node.candidates_weights_quantization_cfg = _create_node_candidates_weights_qc(quant_config,
                                                                                  fw_info,
                                                                                  weight_channel_axis)

    enable_weights_quantization = quant_config.enable_weights_quantization and fw_info.in_kernel_ops(node)
    for qc in node.candidates_weights_quantization_cfg:
        qc.enable_weights_quantization = enable_weights_quantization


def create_node_activation_qc(qc: QuantizationConfig,
                              fw_info: FrameworkInfo) -> NodeActivationQuantizationConfig:
    """
    Create a activations quantization configuration from a QuantizationConfig object.

    Args:
        qc: QuantizationConfig to create the node's config from.
        fw_info: Information about the specific framework the node was created from (e.g., whether or not its
        weights/activations should be quantized)

    Returns:
        Activation quantization configuration of a node.
    """

    activation_quantization_fn = fw_info.activation_quantizer_mapping.get(qc.activation_quantization_method)
    if activation_quantization_fn is None:
        Logger.critical('Unknown quantization method for activations')

    activation_quantization_params_fn = get_activation_quantization_params_fn(qc.activation_quantization_method,
                                                                              qc.activation_error_method)

    return NodeActivationQuantizationConfig(qc,
                                            activation_quantization_fn,
                                            activation_quantization_params_fn)


def create_node_weights_qc(qc: QuantizationConfig,
                           fw_info: FrameworkInfo,
                           weight_channel_axis: int) -> NodeWeightsQuantizationConfig:
    """
    Create a weights quantization configuration from a QuantizationConfig object.

    Args:
        qc: QuantizationConfig to create the node's config from.
        fw_info: Information about the specific framework the node was created from (e.g., whether or not its
        weights/activations should be quantized)
        weight_channel_axis: Axis to quantize a node's kernel when quantizing per-channel.

    Returns:
        Weights quantization configuration of a node.
    """

    weights_quantization_fn = fw_info.weights_quantizer_mapping.get(qc.weights_quantization_method)

    if weights_quantization_fn is None:
        Logger.critical('Unknown quantization method for weights')

    weights_quantization_params_fn = get_weights_quantization_params_fn(qc.weights_quantization_method,
                                                                        qc.weights_error_method)

    return NodeWeightsQuantizationConfig(qc,
                                         weights_quantization_fn,
                                         weights_quantization_params_fn,
                                         weight_channel_axis)


def _create_node_candidates_weights_qc(qc: QuantizationConfig,
                                       fw_info: FrameworkInfo,
                                       weight_channel_axis: int) -> List[NodeWeightsQuantizationConfig]:
    """
    Create a list of candidates of weights quantization configurations for a node.

    Args:
        qc: Quantization configuration the quantization process should follow.
        fw_info: Framework information (e.g., which layers should have their kernels' quantized).
        weight_channel_axis: Output channel index of the node's kernel.

    Returns:
        List of candidates of weights quantization configurations to set for a node.
    """

    candidats = []
    if isinstance(qc, MixedPrecisionQuantizationConfig):
        qc.weights_n_bits.sort(reverse=True)
        for nbits in qc.weights_n_bits:
            single_nbits_qc = copy.deepcopy(qc)
            single_nbits_qc.weights_n_bits = nbits
            candidats.append(create_node_weights_qc(single_nbits_qc, fw_info, weight_channel_axis))
    else:
        candidats.append(create_node_weights_qc(qc, fw_info, weight_channel_axis))

    return candidats



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


# # Layer type constants:
PLACEHOLDER = 'placeholder'
OUTPUT = 'output'

# # Module operation type
CALL_FUNCTION = 'call_function'
CALL_METHOD = 'call_method'

# # Layers attributes constants:
LAYER_NAME = 'name'
USE_BIAS = 'bias'
PADDING = 'padding'
TENSOR_META = 'tensor_meta'
TYPE = 'type'
PAD = 'pad'
VALUE = 'value'
FUNCTIONAL_OP = 'functional_op'
OP_CALL_ARGS = 'op_call_args'
OP_CALL_KWARGS = 'op_call_kwargs'
INPUTS_AS_LIST = 'inputs_as_list'

# # Layers variables names:
KERNEL = 'weight'
BIAS = 'bias'
GAMMA = 'weight'
BETA = 'bias'
MOVING_MEAN = 'running_mean'
MOVING_VARIANCE = 'running_var'
EPSILON = 'eps'

# torch devices
CUDA = 'cuda'
CPU = 'cpu'



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import tensorflow as tf

if tf.__version__ < "2.6":
    from tensorflow.keras.layers import Conv2D, DepthwiseConv2D, Dense, Conv2DTranspose, Reshape, ZeroPadding2D, Dropout, \
        MaxPooling2D, Activation, ReLU, GlobalAveragePooling2D, Add, Multiply, AveragePooling2D, UpSampling2D, InputLayer, \
        Concatenate, Softmax, PReLU, Flatten, Cropping2D, ELU, Dot, LeakyReLU
else:
    from keras.layers import Conv2D, DepthwiseConv2D, Dense, Conv2DTranspose, Reshape, ZeroPadding2D, \
    Dropout, MaxPooling2D, Activation, ReLU, GlobalAveragePooling2D, Add, Multiply, AveragePooling2D, UpSampling2D, \
    InputLayer, Concatenate, Softmax, PReLU, Flatten, Cropping2D, Dot, ELU, LeakyReLU


from model_compression_toolkit.common.defaultdict import DefaultDict
from model_compression_toolkit.common.framework_info import FrameworkInfo, ChannelAxis
from model_compression_toolkit.common.quantization.quantization_config import QuantizationMethod
from model_compression_toolkit.common.quantization.quantizers.kmeans_quantizer import kmeans_quantizer
from model_compression_toolkit.common.quantization.quantizers.lut_kmeans_quantizer import lut_kmeans_quantizer
from model_compression_toolkit.common.quantization.quantizers.uniform_quantizers import power_of_two_quantizer, \
    symmetric_quantizer, uniform_quantizer
from model_compression_toolkit.keras.constants import SOFTMAX, LINEAR, RELU, SWISH, SIGMOID, IDENTITY, TANH, SELU, \
    KERNEL, DEPTHWISE_KERNEL
from model_compression_toolkit.keras.quantizer.fake_quant_builder import power_of_two_quantization, symmetric_quantization, uniform_quantization

"""
Division of Keras layers by how they should be quantized.
KERNEL_OPS: Layers that their coefficients should be quantized.
ACTIVATION: Layers that their activation should be quantized.
NO_QUANTIZATION: Layers that should not be quantized.
"""

KERNEL_OPS = [Conv2D,
              DepthwiseConv2D,
              Dense,
              Conv2DTranspose]

NO_QUANTIZATION = [Reshape,
                   tf.reshape,
                   Flatten,
                   Cropping2D,
                   ZeroPadding2D,
                   Dropout,
                   MaxPooling2D,
                   tf.split,
                   tf.quantization.fake_quant_with_min_max_vars]  # TODO:  replace with marking

ACTIVATION = [Activation,
              ReLU,
              tf.nn.relu,
              tf.nn.relu6,
              tf.nn.leaky_relu,
              Softmax,
              GlobalAveragePooling2D,
              Add,
              Multiply,
              AveragePooling2D,
              UpSampling2D,
              InputLayer,
              Concatenate,
              PReLU,
              ELU,
              tf.nn.silu,
              tf.nn.swish,
              tf.nn.sigmoid,
              tf.nn.tanh,
              tf.nn.relu,
              tf.nn.relu6,
              tf.nn.leaky_relu,
              LeakyReLU,
              tf.nn.softsign,
              tf.nn.gelu,
              tf.nn.elu,
              tf.nn.selu,
              tf.nn.softplus,
              tf.nn.softmax,
              tf.add,
              tf.multiply,
              tf.reduce_mean,
              tf.reduce_min,
              tf.reduce_sum,
              tf.reduce_max,
              tf.image.resize,
              tf.image.crop_and_resize,
              tf.concat,
              Dot]




"""
Map each layer to a list of its' weights attributes that should get quantized.
If a layer that is not listed here is queried, [None] is returned.
"""
KERNEL_ATTRIBUTES = DefaultDict({Conv2D: [KERNEL],
                                 DepthwiseConv2D: [DEPTHWISE_KERNEL],
                                 Dense: [KERNEL],
                                 Conv2DTranspose: [KERNEL]}, lambda: [None])


"""
Map a layer to its kernel's output and input channels indices.
Map's values are tuples of (output_channel_index, input_channel_index).
Default value is returned for layers that are not included.
"""
DEFAULT_CHANNEL_AXIS_DICT = DefaultDict({Conv2D: (3, 2),
                                         DepthwiseConv2D: (2, 2),
                                         Dense: (1, 0),
                                         Conv2DTranspose: (2, 3)}, lambda: (None, None))

"""
Map from an activation function to its min/max output values (if known).
The values are used for tensor min/max values initialization.
"""
ACTIVATION2MINMAX = {SOFTMAX: (0, 1),
                     SIGMOID: (0, 1),
                     LINEAR: (None, None),
                     IDENTITY: (None, None),
                     TANH: (-1, 1),
                     SWISH: (-0.279, None),
                     RELU: (0, None),
                     SELU: (None, None),
                     }

"""
Map from an Keras layer to its min/max output values (if known).
The values are used for tensor min/max values initialization.
"""
LAYER2MINMAX = {Softmax: (0, 1),
                ELU: (-1, None),
                tf.nn.silu: (-0.279, None),
                tf.nn.swish: (-0.279, None),
                tf.nn.sigmoid: (0, 1),
                tf.nn.tanh: (-1, 1),
                tf.nn.relu: (0, None),
                tf.nn.relu6: (0, 6),
                tf.nn.gelu: (-0.17, None),
                tf.nn.elu: (-1, None),
                tf.nn.selu: (-1.76, None),
                tf.nn.softplus: (0, None),
                tf.nn.softmax: (0, 1),
                }
"""
Mapping from a QuantizationMethod to an activation quantizer function.
"""
ACTIVATION_QUANTIZER_MAPPING = {QuantizationMethod.POWER_OF_TWO: power_of_two_quantization,
                                QuantizationMethod.SYMMETRIC: symmetric_quantization,
                                QuantizationMethod.UNIFORM: uniform_quantization}

"""
Mapping from a QuantizationMethod to an weights quantizer function.
"""
WEIGHTS_QUANTIZER_MAPPING = {QuantizationMethod.POWER_OF_TWO: power_of_two_quantizer,
                             QuantizationMethod.SYMMETRIC: symmetric_quantizer,
                             QuantizationMethod.UNIFORM: uniform_quantizer,
                             QuantizationMethod.KMEANS: kmeans_quantizer,
                             QuantizationMethod.LUT_QUANTIZER: lut_kmeans_quantizer}

"""
Output channel index of the model's layers
"""
OUTPUT_CHANNEL_INDEX = ChannelAxis.NHWC

DEFAULT_KERAS_INFO = FrameworkInfo(KERNEL_OPS,
                                   ACTIVATION,
                                   NO_QUANTIZATION,
                                   ACTIVATION_QUANTIZER_MAPPING,
                                   WEIGHTS_QUANTIZER_MAPPING,
                                   DEFAULT_CHANNEL_AXIS_DICT,
                                   ACTIVATION2MINMAX,
                                   LAYER2MINMAX,
                                   KERNEL_ATTRIBUTES,
                                   OUTPUT_CHANNEL_INDEX)



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from copy import deepcopy

import io
import numpy as np
from PIL import Image
from matplotlib.figure import Figure
from tensorboard.compat.proto.attr_value_pb2 import AttrValue
from tensorboard.compat.proto.config_pb2 import RunMetadata
from tensorboard.compat.proto.event_pb2 import Event, TaggedRunMetadata
from tensorboard.compat.proto.graph_pb2 import GraphDef
from tensorboard.compat.proto.node_def_pb2 import NodeDef
from tensorboard.compat.proto.step_stats_pb2 import StepStats, NodeExecStats, DeviceStepStats, AllocatorMemoryUsed
from tensorboard.compat.proto.summary_pb2 import HistogramProto
from tensorboard.compat.proto.summary_pb2 import Summary
from tensorboard.compat.proto.tensor_shape_pb2 import TensorShapeProto
from tensorboard.summary.writer.event_file_writer import EventFileWriter
from typing import List, Any, Dict

from model_compression_toolkit import FrameworkInfo
from model_compression_toolkit.common import Graph, BaseNode
from model_compression_toolkit.common.collectors.statistics_collector import BaseStatsCollector

DEVICE_STEP_STATS = "/device:CPU:0"


def get_node_properties(node_dict_to_log: dict,
                        output_shapes: List[tuple]) -> Dict[str, Any]:
    """
    Create a dictionary with properties for a node to display.

    Args:
        node_dict_to_log: Node's attributes to display.
        output_shapes: Output shapes of the node.

    Returns:
        Dictionary with the node's properties.

    """

    node_properties = {}
    if node_dict_to_log is not None:
        # Add a key-value pair by its type
        for k, v in node_dict_to_log.items():
            if type(v) == int:
                node_properties[str(k)] = AttrValue(i=v)
            elif type(v) == float:
                node_properties[str(k)] = AttrValue(f=v)
            elif type(v) == bool:
                node_properties[str(k)] = AttrValue(b=v)
            else:
                node_properties[str(k)] = AttrValue(s=str(v).encode("utf-8"))

    # Create protobuf for the node's output shapes
    if output_shapes is not None:
        tshape_protos = []
        for output_shape in output_shapes:  # create protobuf for each output shape
            proto_dims_list = []
            for dim in output_shape:
                proto_dims_list.append(TensorShapeProto.Dim(size=dim))
            tshape_proto = TensorShapeProto(dim=proto_dims_list)
            tshape_protos.append(tshape_proto)
        node_properties['_output_shapes'] = AttrValue(list=AttrValue.ListValue(shape=tshape_protos))
    return node_properties


class TensorboardWriter(object):
    """
    Class to log events to display using Tensorboard such as graphs, histograms, images, etc.
    """

    def __init__(self, dir_path: str, fw_info: FrameworkInfo):
        """
        Initialize a TensorboardWriter object.
        
        Args:
            dir_path: Path to save all events to display on Tensorboard.
            fw_info: FrameworkInfo object (needed for computing nodes' weights memory).

        """
        self.dir_path = dir_path
        # we hold EventWriter per tag name, so events can be gathered by tags (like phases during the quantization
        # process).
        self.tag_name_to_event_writer = {}
        self.fw_info = fw_info

    def close(self):
        """

        Close all event-writers the TensorboardWriter holds.
        Should be called at the end of logging process.

        """
        for writer in self.tag_name_to_event_writer.values():
            writer.close()

    def add_histograms(self, graph: Graph, main_tag_name: str):
        """
        Add histograms to display on Tensorboard. All existing histograms in a graph are 
        logged with a tag name main_tag_name.
        
        Args:
            graph: Graph to display all collected histograms it contains.
            main_tag_name: Tag to attach to all histograms.

        """

        def __create_hist_proto(bins: np.ndarray,
                                counts: np.ndarray) -> HistogramProto:
            """
            Create a protobuf for a histogram using collected bins and counts.

            Args:
                bins: Values of histogram bins.
                counts: Counts of histogram values.

            Returns:
                Protobuf of a histogram to display on Tensorboard.
            """

            sum_sq = ((bins * bins) * counts).sum()
            return HistogramProto(min=bins.min(),
                                  max=bins.max(),
                                  num=len(bins),
                                  sum=(bins * counts).sum(),
                                  sum_squares=sum_sq,
                                  bucket_limit=bins.tolist(),
                                  bucket=counts.tolist())

        def __create_histo_event(statistics_collector: BaseStatsCollector):
            """
            Create an event of histogram, and attach it to a list of events outside
            the scope called 'events'.

            Args:
                statistics_collector: Statistics collector to create an event from its histogram.

            """
            if statistics_collector.require_collection():
                if hasattr(statistics_collector, 'hc'):
                    if statistics_collector.hc.is_legal:
                        bins, counts = statistics_collector.hc.get_histogram()
                        if bins is not None and counts is not None:
                            hist = __create_hist_proto(bins[:-1], counts)
                            summary = Summary(value=[Summary.Value(tag=n.name, histo=hist)])
                            events.append(Event(summary=summary))

        events = []
        for n in graph.nodes:
            collector = graph.get_out_stats_collector(n)
            if collector is not None:
                statistics = graph.get_out_stats_collector(n)
                if isinstance(statistics, list):
                    for s in statistics:
                        __create_histo_event(s)
                else:
                    __create_histo_event(statistics)

        # Get the event writer for this tag name
        er = self.__get_event_writer_by_tag_name(main_tag_name)

        for event in events:
            er.add_event(event)
        er.flush()

    def add_graph(self,
                  graph: Graph,
                  main_tag_name: str):
        """
        Add a graph to display on Tensorboard. The graph is tagged with the name main_tag_name.

        Args:
            graph: Graph to display on Tensorboard.
            main_tag_name: Tag to attach to the graph.

        """

        def __get_node_attr(n: BaseNode) -> Dict[str, Any]:
            """
            Create a dictionary to display as the node's attributes.
            The dictionary contains information from node's framework attributes, quantization attributes
            and quantization configuration.

            Args:
                n: Node to create its attributes.

            Returns:
                Dictionary containing attributes to display.
            """

            attr = deepcopy(n.framework_attr)
            if n.quantization_attr is not None:
                attr.update(n.quantization_attr)

            # To log quantization configurations we need to check
            # if they exist at all, as we can log the initial graph,
            # which its nodes do not have configurations yet.
            # Log final config or unified candidates, not both
            if n.final_weights_quantization_cfg is not None:
                attr.update(n.final_weights_quantization_cfg.__dict__)
            elif n.candidates_weights_quantization_cfg is not None:
                attr.update(n.get_unified_candidates_dict())

            if n.activation_quantization_cfg is not None:
                attr.update(n.activation_quantization_cfg.__dict__)
            return attr

        def __get_node_output_dims(n: BaseNode) -> List[tuple]:
            """
            Get node's output shapes. If the first dimension in an output shape is None,
            it means the batch size is dynamic, and it's replaced with -1 to mark it.

            Args:
                n: Node to get its output shapes.

            Returns:
                A list of tuples where each tuple is an output shape of the node.
            """

            # For nodes with an "empty" output shape.
            output_shape = (None,) if n.output_shape == () else n.output_shape

            dims = []
            if isinstance(output_shape, list):
                for o in output_shape:
                    shape_wo_none = (-1,) + o[1:] if o[0] is None else o
                    dims.append(shape_wo_none)
            else:
                dims = [(-1,) + output_shape[1:] if output_shape[0] is None else output_shape]
            return dims

        def __create_node_stats(n: BaseNode):
            """
            Create a NodeExecStats for a node in the graph. A NodeExecStats contains the
            memory and compute time a node requires.

            Args:
                n: Node to create its NodeExecStats.

            Returns:
                A NodeExecStats for a node containing its memory and compute time.

            """

            return NodeExecStats(node_name=n.name,
                                 memory=[AllocatorMemoryUsed(
                                     total_bytes=int(n.get_memory_bytes(self.fw_info))
                                 )])

        graph_def = GraphDef()  # GraphDef to add to Tensorboard

        node_stats = []
        for n in graph.nodes:  # For each node in the graph, we create a NodeDef and connect it to existing NodeDefs
            node_properties = get_node_properties(__get_node_attr(n), __get_node_output_dims(n))
            node_def = NodeDef(attr=node_properties)
            node_def.name = n.name
            node_def.device = n.type.__name__  # For coloring different ops differently
            node_def.op = n.type.__name__
            for e in graph.incoming_edges(n):  # Connect node to its incoming nodes
                i_tensor = f'{e.source_node.name}:{e.source_index}'
                node_def.input.append(i_tensor)
            graph_def.node.extend([node_def])  # Add the node to the graph
            node_stats.append(__create_node_stats(n))

        er = self.__get_event_writer_by_tag_name(main_tag_name)
        event = Event(graph_def=graph_def.SerializeToString())
        er.add_event(event)

        # Logging nodes memory and computation time statistics
        stepstats = RunMetadata(step_stats=StepStats(
            dev_stats=[DeviceStepStats(device=DEVICE_STEP_STATS, node_stats=node_stats)])
        )

        trm = TaggedRunMetadata(tag='Resources', run_metadata=stepstats.SerializeToString())
        event = Event(tagged_run_metadata=trm)
        er.add_event(event)
        er.flush()

    def __get_event_writer_by_tag_name(self,
                                       main_tag_name: str) -> EventFileWriter:
        """
        Retrieve an EventWriter by a tag name from the mapping the TensorboardWriter holds.

        Args:
            main_tag_name: Tag name to retrieve its EventWriter.

        Returns:
            EventFileWriter corresponding to the tag name.
        """

        if main_tag_name in self.tag_name_to_event_writer:  # if an EventWriter already exists, get it
            er = self.tag_name_to_event_writer.get(main_tag_name)
        else:  # if not - create such an EventWriter and save it
            er = EventFileWriter(f'{self.dir_path}/{main_tag_name}')
            self.tag_name_to_event_writer[main_tag_name] = er
        return er

    def add_min_max(self, graph: Graph, main_tag_name: str):
        """
        Add min/max per channel to display on Tensorboard.
        All existing MinMaxPerChannelCollector in a graph are logged with a tag name main_tag_name.

        Args:
            graph: Graph to display all collected MinMaxPerChannelCollectors it contains.
            main_tag_name: Tag to attach to all MinMaxPerChannelCollectors.

        """
        min_events = []
        max_events = []
        for n in graph.nodes:
            collector = graph.get_out_stats_collector(n)
            if collector is not None:
                if hasattr(collector, 'mpcc'):
                    if collector.mpcc.is_legal:
                        mpcc = deepcopy(collector.mpcc)
                        min_pc = mpcc.min_per_channel
                        max_pc = mpcc.max_per_channel
                        for i in range(len(min_pc)):
                            min_i = min_pc[i]
                            max_i = max_pc[i]
                            # use step for channel index as we log the min/max per channel
                            min_events.append(Event(step=i, summary=Summary(
                                value=[Summary.Value(tag=n.name, simple_value=min_i)])))
                            max_events.append(Event(step=i, summary=Summary(
                                value=[Summary.Value(tag=n.name, simple_value=max_i)])))

        # Use a new tag to include both main tag and a 'min_per_channel' tag.
        er = self.__get_event_writer_by_tag_name(main_tag_name + '/min_per_channel')

        for e in min_events:
            er.add_event(e)

        # Use a new tag to include both main tag and a 'max_per_channel' tag.
        er = self.__get_event_writer_by_tag_name(main_tag_name + '/max_per_channel')
        for e in max_events:
            er.add_event(e)

        er.flush()

    def add_mean(self, graph: Graph, main_tag_name: str):
        """
        Add mean per channel to display on Tensorboard.
        All existing MeanCollectors in a graph are logged with a tag name main_tag_name.

        Args:
            graph: Graph to display all collected MeanCollectors it contains.
            main_tag_name: Tag to attach to all MeanCollectors.

        """
        mean_events = []
        for n in graph.nodes:
            collector = graph.get_out_stats_collector(n)
            if collector is not None:
                if hasattr(collector, 'mc'):
                    if collector.mc.is_legal:
                        mc = deepcopy(collector.mc)
                        mean_pc = mc.state
                        for i in range(len(mean_pc)):
                            mean_i = mean_pc[i]
                            # use step for channel index as we log the mean per channel
                            mean_events.append(Event(step=i, summary=Summary(
                                value=[Summary.Value(tag=n.name, simple_value=mean_i)])))

        # Get the event writer for this tag name
        er = self.__get_event_writer_by_tag_name(main_tag_name + '/mean_per_channel')

        for e in mean_events:
            er.add_event(e)

        er.flush()

    def add_all_statistics(self, graph: Graph, main_tag_name: str):
        """
        Add all collected statistics to display on Tensorboard.
        All existing collectors in a graph are logged with a tag name main_tag_name.

        Args:
            graph: Graph to display all collectors it contains.
            main_tag_name: Tag to attach to all collectors display.

        """
        self.add_histograms(graph, main_tag_name)
        self.add_min_max(graph, main_tag_name)
        self.add_mean(graph, main_tag_name)

    def add_figure(self,
                   figure: Figure,
                   figure_tag: str,
                   main_tag_name: str = 'figures'):
        """
        Add matplotlib figure to display. The figure tag is combined from a main_main_tag_name
        and a specific figure_tag for that figure.

        Args:
            figure: Matplotlib figure to display.
            figure_tag: Tag to add to the figure.
            main_tag_name: Main tag which the figure is tagged under.

        """
        figure.canvas.draw()
        data = np.frombuffer(figure.canvas.tostring_rgb(), dtype=np.uint8)
        data = data.reshape(figure.canvas.get_width_height()[::-1] + (3,))

        h, w, c = data.shape
        output = io.BytesIO()
        Image.fromarray(data).save(output, format='PNG')

        img_summary = Summary.Image(height=h, width=w, colorspace=c, encoded_image_string=output.getvalue())
        output.close()

        event = Event(summary=Summary(value=[Summary.Value(tag=figure_tag, image=img_summary)]))

        # Get the event writer for this tag name
        er = self.__get_event_writer_by_tag_name(main_tag_name)
        er.add_event(event)
        er.flush()



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Callable, Any
from model_compression_toolkit.common.defaultdict import DefaultDict

MAX_LSBS_CHANGE_MAP = {8: 4,
                       4: 2,
                       2: 1}


class GradientPTQConfig:
    """
    Configuration to use for quantization with GradientPTQ (experimental).
    """

    def __init__(self,
                 n_iter: int,
                 optimizer: Any,
                 loss: Callable = None,
                 log_function: Callable = None,
                 train_bias: bool = True,
                 lsb_change_per_bit_width: dict = DefaultDict(MAX_LSBS_CHANGE_MAP, lambda: 1)):
        """
        Initialize a GradientPTQConfig.

        Args:
            n_iter (int): Number of iterations to train.
            optimizer (Any): Optimizer to use.
            loss (Callable): the loss to use. should accept 2 lists of tensors. 1st list of quantized tensors, the 2nd list is the float tensors.
            log_function (Callable): Function to log information about the GPTQ process.
            train_bias (bool): Whether to update the bias during the training or not.
            lsb_change_per_bit_width (dict): Whether to update the bias during the training or not.

        """
        self.n_iter = n_iter
        self.optimizer = optimizer
        self.loss = loss
        self.log_function = log_function
        self.train_bias = train_bias
        self.lsb_change_per_bit_width = lsb_change_per_bit_width



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from typing import Any, List


from model_compression_toolkit.common.graph.base_node import BaseNode
from model_compression_toolkit.common.matchers import node_matcher, walk_matcher, edge_matcher


class NodeOperationMatcher(node_matcher.BaseNodeMatcher):
    """
    Class NodeOperationMatcher to check if the layer class of a node matches a specific layer.
    """

    def __init__(self, operation: Any):
        """
        Init for class NodeOperationMathcer.

        Args:
            operation: Which layer to check if matches.
        """

        self.operation = operation

    def apply(self, input_node_object: Any) -> bool:
        """
        Check if input_node_object matches the matcher condition.

        Args:
            input_node_object: Node object to check the matcher on.

        Returns:
            True if input_node_object is the layer the NodeOperationMatcher holds. Otherwise,
            return nothing.
        """

        if input_node_object.type == self.operation:
            return True


class NodeFrameworkAttrMatcher(node_matcher.BaseNodeMatcher):
    """
    Class NodeFrameworkAttrMatcher to check if a node's attribute has a specific value.
    """

    def __init__(self, attr_name: str, attr_value: Any):
        """
        Init a NodeFrameworkAttrMatcher object.

        Args:
            attr_name: Name of node's attribute to check.
            attr_value: Value to check if the attribute is equal to.
        """
        self.attr_name = attr_name
        self.attr_value = attr_value

    def apply(self, input_node_object: Any) -> bool:
        """
        Check if input_node_object has an attribute with the value the NodeFrameworkAttrMatcher
        contains.

        Args:
            input_node_object: Node object to check for its attribute and value.

        Returns:
            True if the node has an attribute with the attribute name and the value that
            were passed during the initialization of NodeFrameworkAttrMatcher.
        """

        if self.attr_name in input_node_object.framework_attr:
            if input_node_object.framework_attr[self.attr_name] == self.attr_value:
                return True


class EdgeMatcher(edge_matcher.BaseEdgeMatcher):
    """
    class EdgeMatcher to check if an edge matches an edge that EdgeMatcher contains.
    """

    def __init__(self, source_matcher: BaseNode, target_matcher: BaseNode):
        """
        Init an EdgeMatcher object.

        Args:
            source_matcher: Source node to match.
            target_matcher: Destination node to match.
        """

        super().__init__(source_matcher, target_matcher)

    def apply(self, input_object: Any) -> bool:
        """
        Check if input_object is a tuple of two nodes and the same nodes that were
        passed during the EdgeMatcher initialization.
        Args:
            input_object: Object to check if equals to the edge EdgeMatcher holds.

        Returns:
            Whether input_object is equal to the edge EdgeMatcher holds or not.
        """

        if isinstance(input_object, tuple) and len(input_object) >= 2:
            return self.source_matcher.apply(input_object[0]) and self.target_matcher.apply(input_object[1])
        else:
            return False


class WalkMatcher(walk_matcher.WalkMatcherList):
    """
    Class WalkMatcher to check if a list of nodes matches another list of nodes.
    """

    def __init__(self, matcher_list: List[BaseNode]):
        """
        Init a WalkMatcher object.

        Args:
            matcher_list: List of nodes to holds for checking.
        """

        super().__init__(matcher_list)

    def apply(self, input_object: Any) -> bool:  # not in use
        """
        Check if a list of nodes matches the list of nodes the WalkMatcher holds.

        Args:
            input_object: Object to check.

        Returns:
            True if input_object matches the list of nodes the WalkMatcher holds.
        """

        pass  # pragma: no cover



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import copy
import os
from functools import partial
from typing import Callable, List, Tuple, Any
from tqdm import tqdm

from model_compression_toolkit import common
from model_compression_toolkit.common import Logger
from model_compression_toolkit.common.gptq.gptq_config import GradientPTQConfig
from model_compression_toolkit.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.common.mixed_precision.kpi import KPI
from model_compression_toolkit.common import FrameworkInfo
from model_compression_toolkit.common.constants import NUM_SAMPLES_CS_TENSORBOARD
from model_compression_toolkit.common.graph.base_graph import Graph
from model_compression_toolkit.common.mixed_precision.bit_width_setter import set_bit_widths

from model_compression_toolkit.common.mixed_precision.mixed_precision_search_facade import search_bit_width
from model_compression_toolkit.common.model_builder_mode import ModelBuilderMode
from model_compression_toolkit.common.network_editors.actions import EditRule
from model_compression_toolkit.common.network_editors.edit_network import edit_network_graph
from model_compression_toolkit.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfig
from model_compression_toolkit.common.quantization.quantization_params_fn_selection import \
    get_activation_quantization_params_fn
from model_compression_toolkit.common.quantization.quantize_graph_weights import quantize_graph_weights
from model_compression_toolkit.common.bias_correction.compute_bias_correction_of_graph import compute_bias_correction_of_graph

from model_compression_toolkit.common.quantization.quantization_analyzer import analyzer_graph
from model_compression_toolkit.common.quantization.quantization_config import DEFAULTCONFIG, QuantizationErrorMethod, \
    QuantizationMethod
from model_compression_toolkit.common.quantization.quantization_config import QuantizationConfig
from model_compression_toolkit.common.quantization.quantization_params_generation.qparams_computation import \
    calculate_quantization_params

from model_compression_toolkit.common.quantization.set_node_quantization_config import set_quantization_configuration_to_graph

from model_compression_toolkit.common.substitutions.apply_substitutions import substitute
from model_compression_toolkit.common.user_info import UserInformation
from model_compression_toolkit.common.model_collector import ModelCollector

from model_compression_toolkit.common.visualization.tensorboard_writer import TensorboardWriter
from model_compression_toolkit.common.bias_correction.apply_bias_correction_to_graph import apply_bias_correction_to_graph


def post_training_quantization(in_model: Any,
                               representative_data_gen: Callable,
                               n_iter: int,
                               quant_config: QuantizationConfig,
                               fw_info: FrameworkInfo,
                               fw_impl: FrameworkImplementation,
                               network_editor: List[EditRule] = [],
                               gptq_config: GradientPTQConfig = None,
                               analyze_similarity: bool = False,
                               target_kpi: KPI = None):
    """
    Quantize a trained model using post-training quantization. The model is quantized using a
    symmetric constraint quantization thresholds (power of two).
    The model is first optimized using several transformations (e.g. BatchNormalization folding to
    preceding layers). Then, using a given dataset, statistics (e.g. min/max, histogram, etc.) are
    being collected for each layer's output (and input, depends on the quantization configuration).
    Thresholds are then being calculated using the collected statistics and the model is quantized
    (both coefficients and activations by default).
    If a gptq configuration is passed, the quantized weights are optimized using knowledge
    distillation by comparing points between the float and quantized models, and minimizing the observed loss.

    Args:
        in_model: Model to quantize.
        representative_data_gen: Dataset used for calibration.
        n_iter: Number of calibration iterations to run.
        quant_config: QuantizationConfig containing parameters of how the model should be quantized. `Default configuration. <https://github.com/sony/model_optimization/blob/21e21c95ca25a31874a5be7af9dd2dd5da8f3a10/model_compression_toolkit/common/quantization/quantization_config.py#L163>`_
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.). `Default Keras info <https://github.com/sony/model_optimization/blob/21e21c95ca25a31874a5be7af9dd2dd5da8f3a10/model_compression_toolkit/keras/default_framework_info.py#L114>`_
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.
        network_editor: List of EditRules. Each EditRule consists of a node filter and an action to change quantization settings of the filtered nodes.
        gptq_config: Configuration for using gradient-based PTQ (e.g. optimizer).
        analyze_similarity: Whether to plot similarity figures within TensorBoard (when logger is enabled) or not.
        target_kpi: KPI to constraint the search of the mixed-precision configuration for the model.

    Returns:
        A quantized model and information the user may need to handle the quantized model.

    """

    tb_w = _init_tensorboard_writer(fw_info)

    tg = _prepare_model_for_quantization(in_model,
                                         representative_data_gen,
                                         network_editor,
                                         n_iter,
                                         quant_config,
                                         fw_info,
                                         tb_w,
                                         fw_impl)


    ######################################
    # Finalize bit widths
    ######################################
    if target_kpi is not None:
        assert isinstance(quant_config, MixedPrecisionQuantizationConfig)
        if quant_config.configuration_overwrite is None:
            bit_widths_config = search_bit_width(tg,
                                                 quant_config,
                                                 fw_info,
                                                 target_kpi,
                                                 partial(fw_impl.get_sensitivity_evaluation_fn,
                                                         representative_data_gen=representative_data_gen,
                                                         fw_info=fw_info))
        else:
            Logger.warning(
                f'Mixed Precision has overwrite bitwidth configuration{quant_config.configuration_overwrite}')
            bit_widths_config = quant_config.configuration_overwrite

    else:
        bit_widths_config = None

    tg = set_bit_widths(quant_config,
                        tg,
                        fw_info,
                        bit_widths_config)

    common.Logger.info(f'Approximated model size (in bytes): {tg.get_memory()}')

    quantized_model, user_info = _quantize_fixed_bit_widths_graph(analyze_similarity,
                                                                  fw_info,
                                                                  gptq_config,
                                                                  representative_data_gen,
                                                                  tb_w,
                                                                  tg,
                                                                  fw_impl)
    user_info.mixed_precision_cfg = bit_widths_config

    return quantized_model, user_info





def _init_tensorboard_writer(fw_info: FrameworkInfo) -> TensorboardWriter:
    """
    Create a TensorBoardWriter object initialized with the logger dir path if it was set,
    or None otherwise.

    Args:
        fw_info: FrameworkInfo object.

    Returns:
        A TensorBoardWriter object.
    """
    tb_w = None
    if common.Logger.LOG_PATH is not None:
        tb_log_dir = os.path.join(os.getcwd(), common.Logger.LOG_PATH, 'tensorboard_logs')
        common.Logger.info(f'To use Tensorboard, please run: tensorboard --logdir {tb_log_dir}')
        tb_w = TensorboardWriter(tb_log_dir, fw_info)
    return tb_w


def _quantize_model(fw_info: FrameworkInfo,
                    tb_w: TensorboardWriter,
                    tg: Graph,
                    fw_impl: FrameworkImplementation) -> Tuple[Any, UserInformation]:
    """
    Quantize graph's weights, and build a quantized Keras model from it.

    Args:
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.).
        tb_w: TensorBoardWriter object to log events.
        tg: A prepared for quantization graph.

    Returns:
        Quantize Keras model, and informat the user may need to use the quantized model.
    """

    quantized_tg = quantize_graph_weights(tg,
                                          fw_info=fw_info,
                                          fw_impl=fw_impl)
    if tb_w is not None:
        tb_w.add_graph(quantized_tg, 'after_quantization')
    ######################################
    # Back2Framework
    ######################################
    # Before building a quantized model, first apply some substitutions.
    quantized_tg = substitute(quantized_tg,
                              fw_impl.get_substitutions_pre_build())
    quantized_model, user_info = fw_impl.model_builder(quantized_tg,
                                                       mode=ModelBuilderMode.QUANTIZED,
                                                       fw_info=fw_info)
    return quantized_model, user_info


def _analyze_similarity(representative_data_gen: Callable,
                        tb_w: TensorboardWriter,
                        tg: Graph,
                        tg_float: Graph):
    """
    Plot the cosine similarity of different points on the graph between the float and quantized
    graphs. Add them to the passed TensorboardWriter object and close all tensorboard writer open
    files.

    Args:
        representative_data_gen: Dataset used for calibration.
        tb_w: TensorBoardWriter object to log events.
        tg: Graph of quantized model.
        tg_float: Graph of float model.

    """
    if tb_w is not None:
        visual = KerasNNVisualizer(tg_float, tg)
        for i in range(NUM_SAMPLES_CS_TENSORBOARD):
            figure = visual.plot_cs_graph(representative_data_gen())
            tb_w.add_figure(figure, f'cosine_similarity_sample_{i}')
        tb_w.close()


def _apply_gptq(gptq_config: GradientPTQConfig,
                representative_data_gen: Callable,
                tb_w: TensorboardWriter,
                tg: Graph,
                tg_bias: Graph,
                fw_info: FrameworkInfo,
                fw_impl: FrameworkImplementation) -> Graph:
    """
    Apply GPTQ to improve accuracy of quantized model.
    Build two models from a graph: A teacher network (float model) and a student network (quantized model).
    and use the dataset generator to pass images through the teacher and student networks to get intermediate
    layers outputs and maximize their similarity.

    Args:
        gptq_config: Configuration for using GPTQ (e.g. optimizer).
        representative_data_gen: Dataset used for calibration.
        tb_w: TensorBoardWriter object to log events.
        tg: Float Reference Graph.
        tg_bias: Graph of quantized model.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.).

    Returns:

    """
    if gptq_config is not None:
        common.Logger.info("Using experimental Gradient Based PTQ: If you encounter an issue "
                           "please file a bug. To disable it, do not pass a gptq configuration.")

        tg_bias = fw_impl.gptq_training(tg,
                                        tg_bias,
                                        representative_data_gen,
                                        gptq_config,
                                        fw_info)

        if tb_w is not None:
            tb_w.add_graph(tg_bias, 'after_gptq')
    return tg_bias


def _quantize_fixed_bit_widths_graph(analyze_similarity: bool,
                                     fw_info: FrameworkInfo,
                                     gptq_config: GradientPTQConfig,
                                     representative_data_gen: Callable,
                                     tb_w: TensorboardWriter,
                                     tg: Graph,
                                     fw_impl: FrameworkImplementation) -> Tuple[Any, UserInformation]:
    """
    Quantize a graph that has final weights candidates quantization configurations.
    Before we quantize the graph weights, we apply GPTQ to get an improved graph.

    Args:
        analyze_similarity: Whether to plot similarity figures within TensorBoard (when logger is enabled) or not.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.)
        gptq_config: Configuration for using GPTQ (e.g. optimizer).
        representative_data_gen: Dataset used for GPTQ fine tuning.
        tb_w: A TensorBoardWriter object initialized with the logger dir path if it was set, or None otherwise.
        tg: Graph to apply GPTQ and to quantize.
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.

    Returns:
        A tuple of the quantized model and an object of UserInformation.

    """
    #############################################
    # Apply Bias Correction
    #############################################
    tg_bias = apply_bias_correction_to_graph(tg,
                                             fw_info=fw_info,
                                             fw_impl=fw_impl)
    if tb_w is not None:
        tb_w.add_graph(tg_bias, 'after_bias_correction')
    #############################################
    # Gradient Based Post Training Quantization
    #############################################
    tg_bias = _apply_gptq(gptq_config,
                          representative_data_gen,
                          tb_w,
                          tg,
                          tg_bias,
                          fw_info,
                          fw_impl)

    tg_float = copy.deepcopy(tg)  # Copy graph before quantization (for similarity analyzer)
    ######################################
    # Final Model Quantization
    ######################################
    quantized_model, user_info = _quantize_model(fw_info,
                                                 tb_w,
                                                 tg_bias,
                                                 fw_impl)
    if analyze_similarity:
        _analyze_similarity(representative_data_gen,
                            tb_w,
                            tg_bias,
                            tg_float)

    return quantized_model, user_info



def _prepare_model_for_quantization(in_model: Any,
                                    representative_data_gen: Callable,
                                    network_editor: List[EditRule] = [],
                                    n_iter: int = 500,
                                    quant_config: QuantizationConfig = DEFAULTCONFIG,
                                    fw_info: FrameworkInfo = None,
                                    tb_w: TensorboardWriter = None,
                                    fw_impl: FrameworkImplementation = None) -> Graph:
    """
    Prepare a trained Keras model for post-training quantization. The model is prepared to be quantized using a
    symmetric constraint quantization thresholds (power of two).
    The model is first read into a graph object and being optimized using several transformations (e.g.
    BatchNormalization folding to preceding layers). Then, using a given dataset, statistics (e.g. min/max,
    histogram, etc.) are being collected for each layer's output (and input, depends on the quantization configuration).
    Thresholds are then being calculated using the collected statistics. Finally, more transformations (based on
    statistics) are applied to increase model's performance.

    Args:
        in_model (Model): Keras model to optimize and prepare for quantization.
        representative_data_gen (Callable): Dataset used for calibration.
        network_editor (List[EditRule]): List of EditRules. Each EditRule consists of a node filter and an action to
        change quantization settings of the filtered nodes.
        n_iter (int): Number of calibration iterations to run.
        quant_config (QuantizationConfig): QuantizationConfig containing parameters of how the model should be
        quantized.
        fw_info (FrameworkInfo): Information needed for quantization about the specific framework (e.g.,
        kernel channels indices, groups of layers by how they should be quantized, etc.)
        tb_w (TensorboardWriter): TensorboardWriter object to use for logging events such as graphs, histograms, etc.
        fw_impl (FrameworkImplementation): FrameworkImplementation object with a specific framework methods implementation.

    Returns:
        Graph object that represents the Keras model, contains thresholds, and ready for quantization.
    """

    ######################################
    # Represent model in a graph
    ######################################
    graph = fw_impl.model_reader(in_model,
                                 representative_data_gen)  # model reading
    graph.set_fw_info(fw_info)


    if tb_w is not None:
        tb_w.add_graph(graph, 'initial_graph')

    ######################################
    # Graph substitution (pre statistics collection)
    ######################################
    transformed_graph = substitute(graph, fw_impl.get_substitutions_pre_statistics_collection())

    if tb_w is not None:
        tb_w.add_graph(transformed_graph, 'pre_statistics_collection_substitutions')

    #########################################
    # Set prior info to nodes
    ##########################################
    for node in transformed_graph.nodes:
        node.prior_info = fw_impl.get_node_prior_info(node=node,
                                                      fw_info=fw_info)


    ######################################
    # Add quantization configurations
    ######################################
    transformed_graph = set_quantization_configuration_to_graph(graph=transformed_graph,
                                                                quant_config=quant_config,
                                                                fw_info=fw_info)

    ######################################
    # Graph marking points
    ######################################
    transformed_graph = substitute(transformed_graph, fw_impl.get_substitutions_marking())

    if tb_w is not None:
        tb_w.add_graph(transformed_graph, 'after_graph_marking')

    ######################################
    # Graph analyzing (attaching statistics collectors)
    ######################################
    analyzer_graph(fw_impl.attach_sc_to_node,
                   transformed_graph,
                   fw_info,
                   quant_config)  # Mark points for statistics collection

    if tb_w is not None:
        tb_w.add_graph(transformed_graph, 'after_analyzer_graph')


    ######################################
    # Statistic collection
    ######################################
    mi = ModelCollector(transformed_graph,
                        fw_impl,
                        fw_info)

    for _ in tqdm(range(n_iter)):
        mi.infer(representative_data_gen())


    ######################################
    # Edit network according to user specific settings
    ######################################
    edit_network_graph(transformed_graph, fw_info, network_editor)

    ######################################
    # Calculate quantization params
    ######################################
    calculate_quantization_params(transformed_graph,
                                  fw_info,
                                  fw_impl=fw_impl)

    if tb_w is not None:
        tb_w.add_graph(transformed_graph, 'thresholds_selection')
        tb_w.add_all_statistics(transformed_graph, 'thresholds_selection')

    ######################################
    # Graph substitution (post statistics collection)
    ######################################
    transformed_graph = substitute(transformed_graph,
                                   fw_impl.get_substitutions_post_statistics_collection(quant_config))


    ######################################
    # Channel equalization
    ######################################
    transformed_graph = substitute(transformed_graph,
                               fw_impl.get_substitutions_channel_equalization(quant_config,
                                                                                  fw_info))

    ######################################
    # Shift Negative Activations
    ######################################
    if quant_config.shift_negative_activation_correction and \
            quant_config.enable_activation_quantization and \
            quant_config.activation_quantization_method is not QuantizationMethod.UNIFORM:
        transformed_graph = fw_impl.shift_negative_correction(transformed_graph,
                                                              quant_config,
                                                              fw_info)
        if tb_w is not None:
            tb_w.add_graph(transformed_graph, 'after_shift_negative_correction')
            tb_w.add_all_statistics(transformed_graph, 'after_shift_negative_correction')

    if tb_w is not None:
        tb_w.add_graph(transformed_graph, 'post_statistics_collection_substitutions')
        tb_w.add_all_statistics(transformed_graph, 'post_statistics_collection_substitutions')

    ########################################################
    # Compute bias correction to nodes' config candidates
    ########################################################
    tg_with_bias = compute_bias_correction_of_graph(transformed_graph,
                                                    fw_info,
                                                    fw_impl)

    if tb_w is not None:
        tb_w.add_graph(tg_with_bias, 'bias_correction_computation')

    for n in tg_with_bias.nodes:
        assert n.final_weights_quantization_cfg is None

    return tg_with_bias




# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Callable

import numpy as np
from scipy import optimize

import model_compression_toolkit.common.quantization.quantization_config as qc
from model_compression_toolkit.common.constants import MIN_THRESHOLD, RANGE_MIN, RANGE_MAX
from model_compression_toolkit.common.quantization.quantization_params_generation.kl_selection import \
    _kl_error_histogram, _kl_error_function
from model_compression_toolkit.common.quantization.quantization_params_generation.lp_selection import \
    _lp_error_histogram
from model_compression_toolkit.common.quantization.quantization_params_generation.mae_selection import \
    _mae_error_histogram
from model_compression_toolkit.common.quantization.quantization_params_generation.mse_selection import \
    _mse_error_histogram
from model_compression_toolkit.common.quantization.quantization_params_generation.qparams_search import \
    qparams_histogram_minimization, kl_uniform_qparams_histogram_minimization, \
    uniform_qparams_selection_per_channel_search, qparams_tensor_minimization
from model_compression_toolkit.common.quantization.quantizers.quantizers_helpers import get_tensor_max, \
    get_tensor_min, uniform_quantize_tensor, get_range_bounds

from model_compression_toolkit.common.similarity_analyzer import compute_mse, compute_mae, compute_lp_norm


def uniform_selection_tensor(tensor_data: np.ndarray,
                             p: int,
                             n_bits: int,
                             per_channel: bool = False,
                             channel_axis: int = 1,
                             n_iter: int = 10,
                             min_threshold: float = MIN_THRESHOLD,
                             quant_error_method: qc.QuantizationErrorMethod = qc.QuantizationErrorMethod.MSE) -> dict:
    """
    Compute the optimal quantization range based on the provided QuantizationErrorMethod
    to uniformly quantize the tensor.
    Different search is applied, depends on the value of the selected QuantizationErrorMethod.

    Args:
        tensor_data: Tensor content as Numpy array.
        p: p-norm to use for the Lp-norm distance.
        n_bits: Number of bits to quantize the tensor.
        per_channel: Whether the quantization should be per-channel or not.
        channel_axis: Output channel index.
        n_iter: Number of iterations to search for the optimal threshold (not used for this method).
        min_threshold: Minimal threshold to use if threshold is too small (not used for this method).
        quant_error_method: an error function to optimize the range parameters' selection accordingly.

    Returns:
        Optimal quantization range to quantize the tensor uniformly.
    """
    tensor_min = get_tensor_min(tensor_data, per_channel, channel_axis)
    tensor_max = get_tensor_max(tensor_data, per_channel, channel_axis)

    if quant_error_method == qc.QuantizationErrorMethod.NOCLIPPING:
        res = tensor_min, tensor_max
    elif quant_error_method == qc.QuantizationErrorMethod.KL:
        # search for KL error is separated because the error method signature is different from the other error methods.
        if per_channel:
            # Using search per-channel wrapper for kl based minimization
            error_function = lambda _x, _y, min_max_range: \
                _kl_error_function(_x, range_min=min_max_range[0], range_max=min_max_range[1], n_bits=n_bits)
            res = uniform_qparams_selection_per_channel_search(tensor_data, tensor_min, tensor_max, channel_axis,
                                                               search_function=lambda _x, _x0, bounds:
                                                               qparams_tensor_minimization(_x, _x0, error_function,
                                                                                           quant_function=lambda min_max_range:
                                                                                           uniform_quantize_tensor(_x,
                                                                                                                   min_max_range[0],
                                                                                                                   min_max_range[1],
                                                                                                                   n_bits),
                                                                                           bounds=bounds))
        else:
            x0 = np.array([tensor_min, tensor_max])
            res = optimize.minimize(fun=lambda min_max_range: _kl_error_function(tensor_data,
                                                                                 range_min=min_max_range[0],
                                                                                 range_max=min_max_range[1],
                                                                                 n_bits=n_bits),
                                    x0=x0,
                                    bounds=get_range_bounds(tensor_min, tensor_max),
                                    method='Nelder-Mead')
            # returned 'x' here is an array with min and max range values
            res = res.x[0], res.x[1]
    else:
        # using error_function wrapper to be able to except range as third parameter for invalid range validation
        _error_function = get_range_selection_tensor_error_function(quant_error_method, p, norm=False)
        error_function = lambda _x, _y, _r: np.inf if _r[1] <= _r[0] else _error_function(_x, _y)
        if per_channel:
            # Using search per-channel wrapper for minimization
            res = uniform_qparams_selection_per_channel_search(tensor_data, tensor_min, tensor_max, channel_axis,
                                                               search_function=lambda _x, _x0, bounds:
                                                               qparams_tensor_minimization(_x, _x0, error_function,
                                                                                           quant_function=lambda min_max_range:
                                                                                           uniform_quantize_tensor(_x,
                                                                                                                   min_max_range[0],
                                                                                                                   min_max_range[1],
                                                                                                                   n_bits),
                                                                                           bounds=bounds))
        else:
            x0 = np.array([tensor_min, tensor_max])
            res = qparams_tensor_minimization(tensor_data, x0, error_function,
                                              quant_function=lambda min_max_range: uniform_quantize_tensor(tensor_data,
                                                                                                           min_max_range[0],
                                                                                                           min_max_range[1],
                                                                                                           n_bits),
                                              bounds=get_range_bounds(tensor_min, tensor_max))
            # returned 'x' here is an array with min and max range values
            res = res.x[0], res.x[1]
    return {RANGE_MIN: res[0],
            RANGE_MAX: res[1]}


def uniform_selection_histogram(bins: np.ndarray,
                                counts: np.ndarray,
                                p: int,
                                n_bits: int,
                                min_value: float,
                                max_value: float,
                                constrained: bool = True,
                                n_iter: int = 10,
                                min_threshold: float = MIN_THRESHOLD,
                                quant_error_method: qc.QuantizationErrorMethod = qc.QuantizationErrorMethod.MSE) -> dict:
    """
    Compute the optimal quantization range based on the provided QuantizationErrorMethod
    to uniformly quantize the histogram.
    Different search is applied, depends on the value of the selected QuantizationErrorMethod.

    Args:
        bins: Bins values of the histogram.
        counts: Bins counts of the histogram.
        p: p-norm to use for the Lp-norm distance (used only for lp threshold selection).
        n_bits: Number of bits to quantize the tensor.
        min_value: Min value (not used for this method).
        max_value: Max value (not used for this method).
        constrained: Whether the threshold should be constrained or not (not used for this method).
        n_iter: Number of iteration ot search for the threshold (not used for this method).
        min_threshold: Minimal threshold to use if threshold is too small (not used for this method).
        quant_error_method: an error function to optimize the range parameters selection accordingly.

    Returns:
        Optimal quantization range to quantize the histogram uniformly.
    """
    tensor_min = np.min(bins[:-1][counts > 0])
    tensor_max = np.max(bins[1:][counts > 0])
    tensor_min_max = np.array([tensor_min, tensor_max])

    if quant_error_method == qc.QuantizationErrorMethod.NOCLIPPING:
        res = tensor_min_max
    elif quant_error_method == qc.QuantizationErrorMethod.KL:
        # search for KL error is separated because the error method signature is different from the other error methods.
        # we pass it as argument to avoid exposing protected package member inside kl_uniform_quantization_loss.
        error_function = _kl_error_histogram
        res = kl_uniform_qparams_histogram_minimization(bins,
                                                        tensor_min_max,
                                                        counts,
                                                        n_bits,
                                                        error_function,
                                                        bounds=get_range_bounds(tensor_min, tensor_max))
        # returned 'x' here is an array with min and max range values
        res = res.x
    else:
        # using error_function wrapper to be able to except range as third parameter for invalid range validation
        _error_function = get_range_selection_histogram_error_function(quant_error_method, p)
        error_function = lambda q_x, _q_count, _x, _counts, _t, _r: np.inf if _r[1] <= _r[0] else \
            _error_function(q_x, _q_count, _x, _counts, _t)
        res = qparams_histogram_minimization(bins,
                                             tensor_min_max,
                                             counts,
                                             error_function,
                                             quant_function=lambda min_max_range:
                                             uniform_quantize_tensor(bins, min_max_range[0], min_max_range[1], n_bits),
                                             bounds=get_range_bounds(tensor_min, tensor_max))
        # returned 'x' here is an array with min and max range values
        res = res.x

    return {RANGE_MIN: res[0], RANGE_MAX: res[1]}


def uniform_no_clipping_selection_min_max(bins: np.ndarray,
                                          counts: np.ndarray,
                                          p: int,
                                          n_bits: int,
                                          min_value: float,
                                          max_value: float,
                                          constrained: bool = False,
                                          n_iter: int = 10,
                                          min_threshold: float = MIN_THRESHOLD,
                                          quant_error_method: qc.QuantizationErrorMethod =
                                          qc.QuantizationErrorMethod.NOCLIPPING) -> dict:
    """
    Gets a quantization rage between min and max numbers.

    Returns:
        A constrained threshold of the min/max values.
    """
    return uniform_selection_histogram(np.asarray([min_value, max_value]),  # histogram with min-max values only
                                       np.asarray([1]),  # passing dummy counts just to make the dimensions work
                                       p,
                                       n_bits,
                                       min_value,
                                       max_value,
                                       constrained,
                                       n_iter,
                                       min_threshold=min_threshold,
                                       quant_error_method=qc.QuantizationErrorMethod.NOCLIPPING)


def get_range_selection_tensor_error_function(quant_error_method: qc.QuantizationErrorMethod,
                                              p: int,
                                              norm: bool = False) -> Callable:
    """
    Returns the error function compatible to the provided threshold method,
    to be used in the threshold optimization search for tensor quantization.
    Args:
        quant_error_method: the requested error function type.
        p: p-norm to use for the Lp-norm distance.
        norm: whether to normalize the error function result.


    Returns: a Callable method that calculates the error between a tensor and a quantized tensor.

    """
    quant_method_error_function_mapping = {
        qc.QuantizationErrorMethod.MSE: lambda x, y: compute_mse(x, y, norm=norm),
        qc.QuantizationErrorMethod.MAE: lambda x, y: compute_mae(x, y, norm=norm),
        qc.QuantizationErrorMethod.LP: lambda x, y: compute_lp_norm(x, y, p=p, norm=norm),
    }

    return quant_method_error_function_mapping[quant_error_method]


def get_range_selection_histogram_error_function(quant_error_method: qc.QuantizationErrorMethod,
                                                 p: int) -> Callable:
    """
    Returns the error function compatible to the provided threshold method,
    to be used in the threshold optimization search for histogram quantization.
    Args:
        quant_error_method: the requested error function type.
        p: p-norm to use for the Lp-norm distance.

    Returns: a Callable method that calculates the error between a tensor and a quantized tensor.

    """
    quant_method_error_function_mapping = {
        qc.QuantizationErrorMethod.MSE: lambda q_bins, q_count, bins, counts, threshold:
            _mse_error_histogram(q_bins, q_count, bins, counts),
        qc.QuantizationErrorMethod.MAE: lambda q_bins, q_count, bins, counts, threshold:
            _mae_error_histogram(q_bins, q_count, bins, counts),
        qc.QuantizationErrorMethod.LP: lambda q_bins, q_count, bins, counts, threshold:
            _lp_error_histogram(q_bins, q_count, bins, counts, p=p),
    }

    return quant_method_error_function_mapping[quant_error_method]



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np
from typing import Tuple, Dict

from model_compression_toolkit import QuantizationMethod
from model_compression_toolkit.common import BaseNode, Graph
from model_compression_toolkit.common.constants import SIGNED
from model_compression_toolkit.common.quantization import quantization_params_generation


def get_activations_qparams(n: BaseNode,
                            graph: Graph) -> Dict[str, float]:
    """
    Compute the activations params for a given node in a graph according to a params function.

    Args:
        n: Node to compute its' activations threshold.
        graph: Graph the node is in.

    Returns:
        The computed activation quantization params.
    """

    out_stats_container = graph.get_out_stats_collector(n)
    bins_values, bins_counts = None, None

    # If the statistics container collected the histogram, we start by filtering outliers using z threshold
    # filtering, and then computing the threshold based on the filtered histogram.
    if out_stats_container.require_collection():
        bins_values, bins_counts = out_stats_container.hc.get_histogram()
        bins_counts = quantization_params_generation.z_score_filter(n.activation_quantization_cfg.z_threshold,
                                                                    bins_values,
                                                                    bins_counts)
    min_value, max_value = out_stats_container.get_min_max_values()

    if n.prior_info.is_output_bounded():
        signed = min_value < 0
    else:
        signed = np.any(bins_values < 0)

    if n.prior_info.is_output_bounded():
        if n.activation_quantization_cfg.activation_quantization_method == QuantizationMethod.POWER_OF_TWO:
            n.activation_quantization_cfg.activation_quantization_params_fn = \
                quantization_params_generation.no_clipping_selection_min_max
        elif n.activation_quantization_cfg.activation_quantization_method == QuantizationMethod.SYMMETRIC:
            n.activation_quantization_cfg.activation_quantization_params_fn = \
                quantization_params_generation.symmetric_no_clipping_selection_min_max
        elif n.activation_quantization_cfg.activation_quantization_method == QuantizationMethod.UNIFORM:
            n.activation_quantization_cfg.activation_quantization_params_fn = \
                quantization_params_generation.uniform_no_clipping_selection_min_max

    activation_params = n.activation_quantization_cfg.activation_quantization_params_fn(bins_values,
                                                                                        bins_counts,
                                                                                        n.activation_quantization_cfg.l_p_value,
                                                                                        n.activation_quantization_cfg.activation_n_bits,
                                                                                        min_value,
                                                                                        max_value,
                                                                                        min_threshold=n.activation_quantization_cfg.min_threshold,
                                                                                        quant_error_method=n.activation_quantization_cfg.activation_error_method)
    activation_params.update({SIGNED: signed})

    return activation_params